from .BaseSerializer import BaseSerializer
from .json.JsonSerializer import JsonSerializer
from .toml.TomlSerializer import TomlSerializer
from .yaml.YamlSerializer import YamlSerializer
from .create_serializer import create_serializer


NEW_LINE = "newline"
SPACE = "space"
LBRACKET = "lbracket"
RBRACKET = "rbracket"
STR = "str"
FIELD_STR = "fieldstr"
POINT = "point"
NUMBER = "number"
BOOL = "bool"
NULL = "null"
ASSIGN = "assign"
EOF = "EOF"

TOKEN_REGEXPS = {
    NEW_LINE: r"\n",
    SPACE: r"\s",
    ASSIGN: "\\=",
    POINT: "\\.",
    NULL: "\\{}",
    LBRACKET: "\\[",
    RBRACKET: "\\]",
    NUMBER: r'([0-9]*[.])?[0-9]+',
    BOOL: r'^(?:tru|fals)e',
    STR: r'"[^"]*"',
    FIELD_STR: r'[A-Za-z0-9\_]*',
    EOF: "EOF"
}


from dataclasses import dataclass


@dataclass
class DTO_TYPES():
    FUNC = "func"
    CODE = "code"
    MODULE = "module"
    CLASS = "class"
    OBJ = "obj"
    DICT = "dict"
    LIST = "list"

@dataclass
class DTO():
    dto_type = "DTO_TYPE"
    name = "name"
    fields = "fields"
    path = "path"
    code = "code"
    global_names = "globals"
    base_class = "class"
    item = "__dto__list_dict_item"



from .BaseSerializer import BaseSerializer
from .json.JsonSerializer import JsonSerializer
from .toml.TomlSerializer import TomlSerializer
from .yaml.YamlSerializer import YamlSerializer


SERAILIZERS_MAP = {
    "json": JsonSerializer,
    "toml": TomlSerializer,
    "yaml": YamlSerializer
}

def create_serializer(exten_name: str) -> BaseSerializer:
    if exten_name in SERAILIZERS_MAP:
        return SERAILIZERS_MAP[exten_name]()
    else:
        return None


import inspect
import re
from types import BuiltinFunctionType, CodeType, GetSetDescriptorType, MappingProxyType, MethodDescriptorType, ModuleType, WrapperDescriptorType

from ... import utils
from ...dto import DTO, DTO_TYPES

from ..BaseSerializer import BaseSerializer
from ...parser.yaml import YamlParser


class YamlSerializer(BaseSerializer):
    __res_str = ""
    __gaps = []
    __gaps_blocked = False
    __parser = None

    def __init__(self):
        super().__init__()
        self.__parser = YamlParser()

    def dumps(self, obj: any) -> str:
        self.__res_str = ""
        self._visit(obj, first_call=True)
        # replace all repeated new lines
        self.__res_str = re.sub(r"(\n)\1{1,}", "\n", self.__res_str)
        return self.__res_str

    def dump(self, obj: any, file_path: str):
        with open(file_path, "w") as file:
            _str = self.dumps(obj)
            file.write(_str)

    def loads(self, source: str) -> any:
        return self.__parser.parse(source)

    def load(self, file_path: str) -> any:
        obj = None
        with open(file_path) as file:
            _str = file.read()
            obj = self.loads(_str)
        return obj

    def _put(self, s: str, gaps: bool = False):
        if gaps and not self.__gaps_blocked:
            self.__res_str += "".join(self.__gaps)
        self.__res_str += s
        self.__gaps_blocked = False

    def _block_gaps(self):
        self.__gaps_blocked = True

    def _push_gap(self, gap: str = "  "):
        self.__gaps.append(gap)

    def _pop_gap(self):
        return self.__gaps.pop()

    def _visit_func_globals(self, func):
        actual_globals = utils.get_actual_func_globals(func)
        self._visit(actual_globals)

    def _visit_func_code(self, _code):
        self._put(f'{DTO.dto_type}: "{DTO_TYPES.CODE}"\n', gaps=True)
        self._put(f'{DTO.fields}:', gaps=True)
        code_dict = utils.get_actual_code_fields(_code)
        self._visit(code_dict)

    def _visit_func(self, func):
        self._put(f'{DTO.dto_type}: "{DTO_TYPES.FUNC}"\n', gaps=True)
        self._put(f'{DTO.name}: "{func.__name__}"\n', gaps=True)
        self._put(f'{DTO.global_names}:', gaps=True)
        self._visit_func_globals(func)
        self._put(f'{DTO.code}: ', gaps=True)
        self._visit(func.__code__)

    def _visit_module(self, module):
        self._put(f'{DTO.dto_type}: "{DTO_TYPES.MODULE}"\n', gaps=True)
        self._put(f'{DTO.name}: "{module.__name__}"\n', gaps=True)
        self._put(f'{DTO.fields}: ', gaps=True)
        if utils.is_std_lib_module(module):
            self._visit(None)
        else:
            module_fields = utils.get_actual_module_fields(module)
            self._visit(module_fields)

    def _visit_class(self, _class):
        self._put(f'{DTO.dto_type}: "{DTO_TYPES.CLASS}"\n', gaps=True)
        self._put(f'{DTO.name}: "{_class.__name__}"\n', gaps=True)
        self._put(f'{DTO.fields}:', gaps=True)
        mems = inspect.getmembers(_class)
        fields_dict = utils.get_actual_class_fields(_class)
        self._visit(fields_dict)

    def _visit_obj(self, obj):
        self._put(f'{DTO.dto_type}: "{DTO_TYPES.OBJ}"\n', gaps=True)
        self._put(f'{DTO.base_class}:', gaps=True)
        self._visit(obj.__class__)
        self._put(f'{DTO.fields}:', gaps=True)
        self._visit(obj.__dict__)

    def _visit_dict(self, _dict: dict):
        self._put(f'{DTO.dto_type}: "{DTO_TYPES.DICT}"\n', gaps=True)
        for item in _dict.items():
            self._put(f'{item[0]}: ', gaps=True)
            # self._push_gap()
            self._visit(item[1])
            self._put("\n")
            # self._pop_gap()

    def _visit_list(self, _list: list):
        if len(_list) >= 1:
            for item in _list:
                self._put("- ", gaps=True)
                if not self._is_primitive_type(item):
                    self._block_gaps()
                self._visit(item, new_line=False)
                self._put("\n")
        else:
            self.__res_str = self.__res_str[:-1]
            self._put("[]")

    def _is_primitive_type(self, obj: any):
        return type(obj) in (int, float, str, bool, bytes)

    def _visit_primitive(self, prim_obj):
        _type = type(prim_obj)
        if _type in (int, float):
            self._put(f'{prim_obj}')
        elif _type == str:
            self._put(f'"{prim_obj}"')
        elif _type == bool:
            val = "true" if prim_obj else "false"
            self._put(f'{val}')
        elif _type == bytes:
            encoded = prim_obj.hex()
            self._put(f'"{encoded}"')

    def _visit(self, obj: any, new_line: bool = True, first_call: bool = False):
        self._push_gap() if not first_call else None

        if self._is_primitive_type(obj):
            self._visit_primitive(obj)
        elif obj == None:
            self._put('null')
        else:
            if len(self.__gaps) >= 1 and new_line:
                self._put("\n")
            if type(obj) == dict:
                self._visit_dict(obj)
            elif type(obj) in (tuple, list):
                self._visit_list(list(obj))
            elif type(obj) == CodeType:
                self._visit_func_code(obj)
            elif type(obj) == ModuleType:
                self._visit_module(obj)
            elif inspect.isclass(obj):
                self._visit_class(obj)
            elif callable(obj):
                self._visit_func(obj)
            elif isinstance(obj, object):
                self._visit_obj(obj)
        self._pop_gap() if not first_call else None



from ast import Module
import imp
import inspect
from types import BuiltinFunctionType, CodeType, GetSetDescriptorType, MappingProxyType, MethodDescriptorType, ModuleType, WrapperDescriptorType
from ..BaseSerializer import BaseSerializer
from ...parser.json.JsonParser import JsonParser
from ...dto import DTO, DTO_TYPES

from ... import utils

class JsonSerializer(BaseSerializer):
    __res_str = ""
    __parser = None

    def __init__(self):
        super().__init__()
        self.__parser = JsonParser()

    def dumps(self, obj: any) -> str:
        self.__res_str = ""
        self._visit(obj)
        return self.__res_str

    def dump(self, obj: any, file_path: str):
        with open(file_path, "w") as file:
            _str = self.dumps(obj)
            file.write(_str)

    def loads(self, source: str) -> any:
        return self.__parser.parse(source)

    def load(self, file_path: str) -> any:
        obj = None
        with open(file_path) as file:
            _str = file.read()
            obj = self.loads(_str)
        return obj

    def _put(self, s: str):
        self.__res_str += s

    def _visit_func_globals(self, func):
        actual_globals = utils.get_actual_func_globals(func)
        self._visit(actual_globals)

    def _visit_func_code(self, _code: CodeType):
        self._put(f'"{DTO.dto_type}": "{DTO_TYPES.CODE}",')
        self._put(f'"{DTO.fields}":')
        code_dict = utils.get_actual_code_fields(_code)
        self._visit(code_dict)

    def _visit_func(self, func):
        self._put(f'"{DTO.dto_type}": "{DTO_TYPES.FUNC}",')
        self._put(f'"{DTO.name}": "{func.__name__}",')
        self._put(f'"{DTO.global_names}": ')
        self._visit_func_globals(func)
        self._put(',')
        self._put(f'"{DTO.code}": ')
        self._visit(func.__code__)

    def _visit_module(self, module):
        self._put(f'"{DTO.dto_type}": "{DTO_TYPES.MODULE}",')
        self._put(f'"{DTO.name}": "{module.__name__}",')
        self._put(f'"{DTO.fields}": ')
        if utils.is_std_lib_module(module):
            self._visit(None)
        else:
            module_fields = utils.get_actual_module_fields(module)
            self._visit(module_fields)

    def _visit_class(self, _class):
        self._put(f'"{DTO.dto_type}": "{DTO_TYPES.CLASS}",')
        self._put(f'"{DTO.name}": "{_class.__name__}",')
        self._put(f'"{DTO.fields}": ')
        fields_dict = utils.get_actual_class_fields(_class)
        self._visit(fields_dict)

    def _visit_obj(self, obj):
        # print(obj)
        self._put(f'"{DTO.dto_type}": "{DTO_TYPES.OBJ}",')
        self._put(f'"{DTO.base_class}": ')
        self._visit(obj.__class__)
        self._put(",")
        self._put(f'"{DTO.fields}": ')
        self._visit(obj.__dict__)

    def _visit_dict(self, _dict: dict):
        self._put(f'"{DTO.dto_type}": "{DTO_TYPES.DICT}"')
        if len(_dict.items()) >= 1:
            self._put(",")
        is_first = True
        for item in _dict.items():
            if not is_first:
                self._put(',')
            self._put(f'"{item[0]}": ')
            self._visit(item[1])
            is_first = False

    def _visit_primitive(self, prim_obj):
        _type = type(prim_obj)
        if _type in (int, float):
            self._put(f'{prim_obj}')
        elif _type == str:
            self._put(f'"{prim_obj}"')
        elif _type == bool:
            val = "true" if prim_obj else "false"
            self._put(f'{val}')
        elif _type in (list, tuple):
            self._put('[')
            for i, obj in enumerate(prim_obj):
                if i != 0:
                    self._put(",")
                self._visit(obj)
            self._put(']')
        elif _type == bytes:
            encoded = prim_obj.hex()
            self._put(f'"{encoded}"')

    def _visit(self, obj):
        if type(obj) in (int, float, str, bool, bytes, tuple, list):
            self._visit_primitive(obj)
        elif obj == None:
            self._put('null')
        else:
            self._put('{')
            if type(obj) == dict:
                self._visit_dict(obj)
            elif type(obj) == CodeType:
                self._visit_func_code(obj)
            elif type(obj) == ModuleType:
                self._visit_module(obj)
            elif inspect.isclass(obj):
                self._visit_class(obj)
            elif callable(obj):
                self._visit_func(obj)
            elif isinstance(obj, object):
                self._visit_obj(obj)
            self._put('}')



import inspect
import re
from types import BuiltinFunctionType, CodeType, GetSetDescriptorType, MappingProxyType, MethodDescriptorType, ModuleType, WrapperDescriptorType
from ... import utils
from ...parser.toml.TomlParser import TomlParser
from ...dto import DTO, DTO_TYPES
from ..BaseSerializer import BaseSerializer


class TomlSerializer(BaseSerializer):
    __res_str = ""
    __container_names = []
    __parser = None

    def __init__(self):
        super().__init__()
        self.__parser = TomlParser()

    def dumps(self, obj: any) -> str:
        self.__res_str = ""
        self._visit(obj)
        # replace all repeated new lines
        self.__res_str = re.sub(r"(\n)\1{2,}", "\n\n", self.__res_str)
        return self.__res_str

    def dump(self, obj: any, file_path: str):
        with open(file_path, "w") as file:
            _str = self.dumps(obj)
            file.write(_str)

    def loads(self, source: str) -> any:
        return self.__parser.parse(source)

    def load(self, file_path: str) -> any:
        obj = None
        with open(file_path) as file:
            _str = file.read()
            obj = self.loads(_str)
        return obj

    def _put(self, s: str):
        self.__res_str += s

    def _push_name(self, s: str):
        self.__container_names.append(s)

    def _pop_name(self) -> str:
        return self.__container_names.pop()

    def _get_concat_name(self) -> str:
        return (".".join(self.__container_names))[1:]

    def _is_primitive_type(self, obj: any) -> bool:
        _type = type(obj)
        if _type in (int, float, str, bool, bytes) or obj == None:
            return True
        elif _type in (tuple, list):
            if len(obj) >= 1:
                # return all(self._is_primitive_type(obj[i])==True for i in range(len(obj)))
                return False
            else:
                return True
        return False

    def _divide_dict_by_primitive(self, _dict: dict) -> tuple:
        prim_dict = {}
        complex_dict = {}
        for item in _dict.items():
            if self._is_primitive_type(item[1]):
                prim_dict.update({item[0]: item[1]})
            else:
                complex_dict.update({item[0]: item[1]})
        return prim_dict, complex_dict

    def _divide_list_by_primitive(self, _list: list) -> tuple:
        prim_list = []
        complex_list = []
        for item in _list:
            if self._is_primitive_type(item):
                prim_list.append(item)
            else:
                complex_list.append(item)
        return prim_list, complex_list

    def _visit_func_globals(self, func):
        actual_globals = utils.get_actual_func_globals(func)
        self._visit(actual_globals, DTO.global_names)

    def _visit_func_code(self, func):
        self._put(f'{DTO.dto_type} = "{DTO_TYPES.CODE}"\n\n')
        code_dict = utils.get_actual_code_fields(func)
        self._visit(code_dict, DTO.fields)

    def _visit_func(self, func):
        self._put(f'{DTO.dto_type} = "{DTO_TYPES.FUNC}"\n')
        self._put(f'{DTO.name} = "{func.__name__}"\n\n')
        self._visit_func_globals(func)
        self._visit(func.__code__, DTO.code)
        # exit()

    def _visit_module(self, module):
        self._put(f'{DTO.dto_type} = "{DTO_TYPES.MODULE}"\n')
        self._put(f'{DTO.name} = "{module.__name__}"\n')
        if utils.is_std_lib_module(module):
            self._put(f'{DTO.fields} = ')
            self._visit(None)
            self._put("\n")
        else:
            self._put("\n")
            module_fields = utils.get_actual_module_fields(module)
            self._visit(module_fields, DTO.fields)

    def _visit_class(self, _class):
        self._put(f'{DTO.dto_type} = "{DTO_TYPES.CLASS}"\n')
        self._put(f'{DTO.name} = "{_class.__name__}"\n\n')
        # self._put(f'"{DTO.fields}": ')
        fields_dict = utils.get_actual_class_fields(_class)
        self._visit(fields_dict, DTO.fields)

    def _visit_obj(self, obj):
        # print(obj)
        self._put(f'{DTO.dto_type} =  "{DTO_TYPES.OBJ}"\n\n')
        self._visit(obj.__class__, DTO.base_class)
        self._visit(obj.__dict__, DTO.fields)

    def _visit_dict(self, _dict: dict):
        # sorting dict to set primitive fields at the begining
        prim_dict, complex_dict = self._divide_dict_by_primitive(_dict)

        self._put(f'{DTO.dto_type} = "{DTO_TYPES.DICT}"\n')

        # print(prim_dict, complex_dict)
        # exit()

        for prim in prim_dict.items():
            self._put(f'{prim[0]} = ')
            self._visit(prim[1])
            self._put("\n")

        self._put("\n")

        for comp in complex_dict.items():
            self._visit(comp[1], comp[0])
            self._put("\n")

    def _visit_primitive(self, prim_obj):
        _type = type(prim_obj)
        if _type in (int, float):
            self._put(f'{prim_obj}')
        elif _type == str:
            self._put(f'"{prim_obj}"')
        elif _type == bool:
            val = "true" if prim_obj else "false"
            self._put(f'{val}')
        elif prim_obj == None:
            self._put("{}")
        elif type(prim_obj) in (tuple, list) and len(prim_obj) == 0:
            self._put("[]")
            return
        elif _type == bytes:
            encoded = prim_obj.hex()
            self._put(f'"{encoded}"')

    def _visit_list(self, _list: any):
        prim_list, complex_list = self._divide_list_by_primitive(_list)
        self._put(f'{DTO.dto_type} = "{DTO_TYPES.LIST}"\n')
        i = 0
        for prim in prim_list:
            self._put(f'{DTO.item}{i} = ')
            self._visit(prim)
            self._put("\n")
            i += 1
        self._put("\n")
        for comp in complex_list:
            self._visit(comp, f'{DTO.item}{i}')
            self._put("\n")
            i += 1

    def _visit_complex(self, comp_obj: any, container_name: str):
        self._push_name(container_name)
        name = self._get_concat_name()
        if len(self.__container_names) > 1:
            self._put(f'[{name}]\n')
        if type(comp_obj) == dict:
            self._visit_dict(comp_obj)
        elif type(comp_obj) in (tuple, list):
            self._visit_list(comp_obj)
        elif type(comp_obj) == ModuleType:
            self._visit_module(comp_obj)
        elif inspect.isclass(comp_obj):
            self._visit_class(comp_obj)
        elif type(comp_obj) == CodeType:
            self._visit_func_code(comp_obj)
        elif callable(comp_obj):
            self._visit_func(comp_obj)
        elif isinstance(comp_obj, object):
            self._visit_obj(comp_obj)

        self._pop_name()

    def _visit(self, obj, new_name: str = ""):
        if self._is_primitive_type(obj):
            self._visit_primitive(obj)
        else:
            self._visit_complex(obj, new_name)



from abc import ABC, abstractmethod


class BaseSerializer(ABC):
    @abstractmethod
    def dumps(self, obj: any) -> str:
        """ this method is too abstract to understand. """
    @abstractmethod
    def dump(self, obj: any, file_path: str):
        """ this method is too abstract to understand. """
    @abstractmethod
    def loads(self, source: str) -> any:
        """ this method is too abstract to understand. """
    @abstractmethod
    def load(self, file_path: str) -> any:
        """ this method is too abstract to understand. """


import sys

sys.path.append(sys.path[0] + "/.")

# sys.path.append("..")

from serializer import create_serializer
# from cli.CLI import CLI

exit()
def main():
    cli = CLI()
    args = cli.parse_args()
    cli.exec_convert(args)



if __name__ == "__main__":
    main()


LBRACE = "lbrace"
RBRACE = "rbrace"
LBRACKET = "lbracket"
RBRACKET = "rbracket"
NULL = "null"
STR = "str"
NUMBER = "number"
BOOL = "bool"
COLON = "colon"
COMMA = "comma"
EOF = "EOF"

TOKEN_REGEXPS = {
    LBRACE: r"{",
    RBRACE: r"}",
    LBRACKET: "\\[",
    RBRACKET: "\\]",
    NULL: r'null',
    STR: r'"[^"]*"',
    NUMBER: r'([0-9]*[.])?[0-9]+',
    BOOL: r'^(?:tru|fals)e',
    COLON: r":",
    COMMA: r",",
    EOF: "EOF"
}


import imp
import inspect
import re
from types import CodeType, FunctionType, ModuleType

from ...dto import DTO, DTO_TYPES
from . import toml_tokens
TOKEN_TYPES = toml_tokens

class TomlParser():

    __tokens: list = []
    __container_names = []

    def _eat(self, token_types: tuple) -> tuple:
        if len(self.__tokens):
            if self.__tokens[0][0] in token_types:
                return self.__tokens.pop(0)
        return ("", "")

    def _head_token(self) -> tuple:
        if len(self.__tokens):
            return self.__tokens[0]

    def _lex(self, source: str) -> list:
        tokens = []
        while len(source) > 0:
            for token in TOKEN_TYPES.TOKEN_REGEXPS.items():
                if token[0] == TOKEN_TYPES.EOF:
                    continue
                regexp_res = re.match(token[1], source)
                if regexp_res and regexp_res.start() == 0 and len(regexp_res.group(0)):
                    source = str(
                        source[regexp_res.end()-regexp_res.start():])
                    if token[0] in (TOKEN_TYPES.STR, TOKEN_TYPES.FIELD_STR):
                        string = regexp_res.group(0)
                        tokens.append((token[0], string.replace('"', "")))
                    elif token[0] == TOKEN_TYPES.NUMBER:
                        num = regexp_res.group(0)
                        num = float(num) if "." in num else int(num)
                        tokens.append((token[0], num))
                    elif token[0] == TOKEN_TYPES.BOOL:
                        _bool = regexp_res.group(0)
                        res = True if _bool == "true" else False
                        tokens.append((token[0], res))
                    else:
                        tokens.append((token[0],))
        tokens.append((TOKEN_TYPES.EOF,))
        return tokens

    def skip_newlines(self):
        while self._head_token()[0] == TOKEN_TYPES.NEW_LINE:
            self._eat(TOKEN_TYPES.NEW_LINE)

    def _push_name(self, s: str):
        self.__container_names.append(s)

    def _pop_name(self) -> str:
        return self.__container_names.pop()

    def _parse_container_names(self, no_eat: bool = False) -> list:
        tokens_copy = [t for t in self.__tokens]
        self._eat(TOKEN_TYPES.LBRACKET)
        names = []
        while self._head_token()[0] != TOKEN_TYPES.RBRACKET:
            token = self._eat((TOKEN_TYPES.FIELD_STR, TOKEN_TYPES.POINT))
            if token[0] == TOKEN_TYPES.FIELD_STR:
                names.append(token[1])
        self._eat(TOKEN_TYPES.RBRACKET)
        self._eat(TOKEN_TYPES.NEW_LINE)
        if no_eat:
            self.__tokens = tokens_copy
        return names

    def _skip_field_name(self) -> str:
        field_key = self._eat(TOKEN_TYPES.FIELD_STR)
        self._eat(TOKEN_TYPES.ASSIGN)
        return field_key[1]

    def _parse_func_code(self) -> CodeType:
        # fields
        self._skip_field_name()
        code_dict = self._parse()
        code_dict["co_consts"] = sorted(code_dict["co_consts"], key = lambda el: str(el).find("<"))

        func_code = CodeType(
            int(code_dict["co_argcount"]),
            int(code_dict["co_posonlyargcount"]),
            int(code_dict["co_kwonlyargcount"]),
            int(code_dict["co_nlocals"]),
            int(code_dict["co_stacksize"]),
            int(code_dict["co_flags"]),
            bytes.fromhex(code_dict["co_code"]),
            tuple(code_dict["co_consts"]),
            tuple(code_dict["co_names"]),
            tuple(code_dict["co_varnames"]),
            str(code_dict["co_filename"]),
            str(code_dict["co_name"]),
            int(code_dict["co_firstlineno"]),
            bytes.fromhex(code_dict["co_lnotab"]),
            tuple(code_dict["co_freevars"]),
            tuple(code_dict["co_cellvars"]),
        )
        return func_code

    def _parse_func(self) -> any:
        # name
        self._skip_field_name()
        func_name = self._parse()
        self.skip_newlines()
        # globals
        self._skip_field_name()
        func_globals = self._parse()
        self.skip_newlines()
        # code
        self._skip_field_name()
        func_code = self._parse()
        self.skip_newlines()

        # print(func_code)

        # print("_________")
        # for c in func_code.items():
        #     print(c)
        # exit()

        func = FunctionType(func_code, func_globals, func_name)
        func.__globals__["__builtins__"] = __import__("builtins")
        return func

    # def _is_sub_list(self, extern_list: list, internal_list: list) -> bool:
    #     ex_len, in_len = len(extern_list), len(internal_list)

    def _parse_module(self) -> ModuleType:
        # name
        self._skip_field_name()
        module_name = self._parse()
        # fields
        self._skip_field_name()
        module_fields = self._parse()

        module = None
        # first case mean that module is in python std or built in
        if module_fields == None:
            module = __import__(module_name)
        else: 
            module = imp.new_module(module_name)
            for field in module_fields.items():
                setattr(module,field[0],field[1])
        return module

    def _parse_class(self) -> type:
        # name
        self._skip_field_name()
        class_name = self._parse()
        # fields
        self._skip_field_name()
        class_members_dict = self._parse()

        class_bases = (object,)
        if "__bases__" in class_members_dict:
            class_bases = tuple(class_members_dict["__bases__"])
        return type(class_name, class_bases, class_members_dict)

    def _parse_obj(self) -> object:
        # class
        self._skip_field_name()
        _class = self._parse()
        # fields
        self._skip_field_name()
        fields_dict = self._parse_dto()

        class_init = _class.__init__
        if callable(class_init):
            if class_init.__class__.__name__ == "function":
                delattr(_class, "__init__")
        obj = _class()
        obj.__init__ = class_init
        obj.__dict__ = fields_dict
        return obj

    def _parse_dict(self):
        # print(self.__container_names)
        _dict = {}
        while self._head_token()[0] not in (TOKEN_TYPES.LBRACKET, TOKEN_TYPES.EOF):
            key = self._skip_field_name()
            value = self._parse()
            _dict.update({key: value})
            self.skip_newlines()
        if self._head_token()[0] != TOKEN_TYPES.EOF:
            while self._head_token()[0] != TOKEN_TYPES.EOF:
                next_names = self._parse_container_names(no_eat=True)
                if next_names[:-1] == self.__container_names or len(self.__container_names) == 0:
                    field_name = next_names[-1:][0]
                    value = self._parse_dto()
                    _dict.update({field_name: value})
                    if self._head_token()[0] == TOKEN_TYPES.EOF:
                        break
                    self.skip_newlines()
                    next_names = self._parse_container_names(no_eat=True)
                else:
                    break
        # print(self.__container_names,"end")
        return _dict

    def _parse_list(self) -> list:
        _dict = self._parse_dict()
        return list(_dict.values())

    def _parse_primitive(self) -> any:
        token_type = self._head_token()[0]
        res = None
        if token_type == TOKEN_TYPES.NUMBER:
            res = self._eat(TOKEN_TYPES.NUMBER)[1]
        elif token_type == TOKEN_TYPES.STR:
            res = self._eat(TOKEN_TYPES.STR)[1]
        elif token_type == TOKEN_TYPES.LBRACKET:
            self._eat(TOKEN_TYPES.LBRACKET)
            self._eat(TOKEN_TYPES.RBRACKET)
            res = []
        elif token_type == TOKEN_TYPES.NULL:
            self._eat(TOKEN_TYPES.NULL)
            res = None
        elif token_type == TOKEN_TYPES.BOOL:
            res = self._eat(TOKEN_TYPES.BOOL)[1]
        self._eat(TOKEN_TYPES.NEW_LINE)
        return res

    def _parse_dto(self, global_scope: bool = False):
        self.skip_newlines()
        if not global_scope:
            names: list = self._parse_container_names()
            if len(names) >= 1:
                self._push_name(names[len(names)-1])

        dto_type_key = self._eat(TOKEN_TYPES.FIELD_STR)
        self._eat(TOKEN_TYPES.ASSIGN)
        dto_type_value = self._eat(TOKEN_TYPES.STR)
        self._eat(TOKEN_TYPES.NEW_LINE)

        res_dto = None

        if dto_type_key[1] == DTO.dto_type:
            if dto_type_value[1] == DTO_TYPES.DICT:
                res_dto = self._parse_dict()
            elif dto_type_value[1] == DTO_TYPES.LIST:
                res_dto = self._parse_list()
            elif dto_type_value[1] == DTO_TYPES.FUNC:
                res_dto = self._parse_func()
            elif dto_type_value[1] == DTO_TYPES.CODE:
                res_dto = self._parse_func_code()
            elif dto_type_value[1] == DTO_TYPES.MODULE:
                res_dto = self._parse_module()
            elif dto_type_value[1] == DTO_TYPES.CLASS:
                res_dto = self._parse_class()
            elif dto_type_value[1] == DTO_TYPES.OBJ:
                res_dto = self._parse_obj()

        if not global_scope:
            self._pop_name()

        return res_dto

    def _parse(self) -> any:
        self.skip_newlines()
        head_token_type = self._head_token()[0]
        if head_token_type == TOKEN_TYPES.FIELD_STR:
            return self._parse_dto(global_scope=True)
        elif head_token_type == TOKEN_TYPES.LBRACKET:
            # check below need to avoid empty array during parsing in parse_dto method
            if self.__tokens[1][0] != TOKEN_TYPES.RBRACKET:
                return self._parse_dto()
        return self._parse_primitive()

    def parse(self, s: str) -> any:
        self.__tokens = self._lex(s)
        # remove all space tokens
        self.__tokens = list(filter(
            lambda token: token[0] != TOKEN_TYPES.SPACE,
            self.__tokens
        ))
        return self._parse()





from setuptools import setup, find_packages


setup(
    name="pekutko-serializer",
    version="0.0.8",
    description="Dmitry Pekutko, gruop 053504, format cenverter (Serializer)",
    author="Dmitry Pekutko",
    author_email="dkefir03@gmail.com",
    python_requires='>3.8.0',
    packages=find_packages(),
    url="https://github.com/DimaPekutko/ISP-2022-053504/",
)